#!/usr/bin/env python3
"""
Add a STAT table to a weight only variable font.

This script can also add STAT tables to a variable font family which
consists of two fonts, one for Roman, the other for Italic.
Both of these fonts must also only contain a weight axis.

For variable fonts with multiple axes, write a python script which
uses fontTools.otlLib.builder.buildStatTable e.g
https://github.com/googlefonts/literata/blob/master/sources/gen_stat.py

The generated STAT tables use format 2 Axis Values. These are needed in
order for Indesign to work.

Special mention to Thomas Linard for reviewing the output of this script.


Usage:

Single family:
gftools fix-vf-meta FontFamily[wght].ttf

Roman + Italic family:
gftools fix-vf-meta FontFamily[wght].ttf FontFamily-Italic[wght].ttf
"""
from fontTools.otlLib.builder import buildStatTable
from fontTools.ttLib import TTFont
import argparse


WGHT = {
    100: "Thin",
    200: "ExtraLight",
    300: "Light",
    400: "Regular",
    500: "Medium",
    600: "SemiBold",
    700: "Bold",
    800: "ExtraBold",
    900: "Black",
    1000: "ExtraBlack",
}


def font_is_italic(ttfont):
    stylename = ttfont["name"].getName(2, 3, 1, 0x409).toUnicode()
    return True if "Italic" in stylename else False


def font_has_mac_names(ttfont):
    for record in ttfont['name'].names:
        if record.platformID == 1:
            return True
    return False


def build_stat(roman_font, italic_font=None):
    roman_wght_axis = dict(
        tag="wght",
        name="Weight",
        values=build_axis_values(roman_font),
    )
    roman_axes = [roman_wght_axis]
    if italic_font:
    # We need to create a new Italic axis in the Roman font
        roman_axes.append(
            dict(
                tag="ital",
                name="Italic",
                values=[
                    dict(
                        name="Roman",
                        flags=2,
                        value=0.0,
                        linkedValue=1.0,
                    )
                    
                ]
            )
        )
        italic_wght_axis = dict(
            tag="wght",
            name="Weight",
            values=build_axis_values(italic_font),
        )
        italic_axes = [italic_wght_axis]
        italic_axes.append(
            dict(
                tag="ital",
                name="Italic",
                values=[
                    dict(
                        name="Italic",
                        value=1.0,
                    )
                ]
            )
        )
        buildStatTable(italic_font, italic_axes)
    buildStatTable(roman_font, roman_axes)


def build_axis_values(ttfont):
    results = []
    nametable = ttfont['name']
    instances = ttfont['fvar'].instances
    has_bold = any([True for i in instances if i.coordinates['wght'] == 700])
    for instance in instances:
        wght_val = instance.coordinates["wght"]
        desired_inst_info = WGHT[wght_val]
        name = nametable.getName(
            instance.subfamilyNameID,
            3,
            1,
            1033
        ).toUnicode()
        name = name.replace("Italic", "").strip()
        if name == "":
            name = "Regular"
        inst = {
            "name": name,
            "nominalValue": wght_val,
        }
        if inst["nominalValue"] == 400:
            inst["flags"] = 0x2
        results.append(inst)

    # Dynamically generate rangeMinValues and rangeMaxValues
    entries = [results[0]["nominalValue"]] + \
              [i["nominalValue"] for i in results] + \
              [results[-1]["nominalValue"]]
    for i, entry in enumerate(results):
        entry["rangeMinValue"] = (entries[i] + entries[i+1]) / 2
        entry["rangeMaxValue"] = (entries[i+1] + entries[i+2]) / 2

    # Format 2 doesn't support linkedValues so we have to append another
    # Axis Value (format 3) for Reg which does support linkedValues
    if has_bold:
        inst = {
            "name": "Regular",
            "value": 400,
            "flags": 0x2,
            "linkedValue": 700
        }
        results.append(inst)
    return results


def update_nametable(ttfont):
    """
    - Add nameID 25
    - Update fvar instance names and add fvar instance postscript names
    """
    is_italic = font_is_italic(ttfont)
    has_mac_names = font_has_mac_names(ttfont)

    # Add nameID 25
    # https://docs.microsoft.com/en-us/typography/opentype/spec/name#name-ids
    vf_ps_name = _add_nameid_25(ttfont, is_italic, has_mac_names)

    # Update fvar instances
    instances = ttfont["fvar"].instances
    for inst in instances:
        wght_val = inst.coordinates["wght"]
        if wght_val not in WGHT:
            raise ValueError(f"Unsupported wght coord '{wght_val}'. Coord "
                              "needs to be in {WGHT.keys()}")

        # Update instance subfamilyNameID
        wght_name = WGHT[wght_val]
        inst_name = wght_name
        if is_italic:
            inst_name = f"{inst_name} Italic"
            inst_name = inst_name.replace("Regular Italic", "Italic")
        ttfont['name'].setName(inst_name, inst.subfamilyNameID, 3, 1, 0x409)
        if has_mac_names:
            ttfont['name'].setName(inst_name, inst.subfamilyNameID, 1, 0, 0)

        # Add instance psName
        ps_name = f"{vf_ps_name}-{wght_name}"
        ps_name_id = ttfont['name'].addName(ps_name)
        inst.postscriptNameID = ps_name_id


def _add_nameid_25(ttfont, is_italic, has_mac_names):
    name = ttfont['name'].getName(16, 3, 1, 1033) or \
        ttfont['name'].getName(1, 3, 1, 1033)
    name = name.toUnicode().replace(" ", "")
    if is_italic:
        name = f"{name}Italic"
    else:
        name = f"{name}Roman"
    ttfont['name'].setName(name, 25, 3, 1, 1033)
    if has_mac_names:
        ttfont['name'].setName(name, 25, 1, 0, 0)
    return name


def main():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=__doc__
    )
    parser.add_argument("fonts", nargs="+", help=(
            "Paths to font files. Fonts must be part of the same family."
        )
    )
    args = parser.parse_args()
    fonts = args.fonts

    # This monstrosity exists so we don't break the v1 api.
    italic_font = None
    if len(fonts) > 2:
        raise Exception(
            "Can only add STAT tables to a max of two fonts. "
            "Run gftools fix-vf-meta --help for usage instructions"
        )
    elif len(fonts) == 2:
        if "Italic" in fonts[0]:
            italic_font = TTFont(fonts[0])
            roman_font = TTFont(fonts[1])
        elif "Italic" in fonts[1]:
            italic_font = TTFont(fonts[1])
            roman_font = TTFont(fonts[0])
        else:
            raise Exception("No Italic font found!")
    else:
        roman_font = TTFont(fonts[0])
    update_nametable(roman_font)
    if italic_font:
        update_nametable(italic_font)
    build_stat(roman_font, italic_font)
    roman_font.save(roman_font.reader.file.name + ".fix")
    if italic_font:
        italic_font.save(italic_font.reader.file.name + ".fix")


if __name__ == "__main__":
    main()




#!/usr/bin/env python3
"""
gftools gen-stat

Generate a STAT table for each font in a variable font family
using the GF axis registry.

Usage:

# Standard usage. Fonts will have ".fix" appended to their filenames
gftools gen-stat font1.ttf --axis-order wdth wght

# Output fonts to a dir
gftools gen-stat font1.ttf font2.ttf --axis-order wdth wght --out ~/Desktop/out

# Overwrite input fonts
gftools gen-stat font1.ttf font2.ttf --axis-order wdth wght --inplace

# Overide which axis values are elided
gftools gen-stat font.ttf --elided-values wght=400 --axis-order wdth wght

"""
from fontTools.ttLib import TTFont
from gftools.stat import gen_stat_tables
from gftools.axisreg import axis_registry
import argparse
import os


def parse_elided_values(string):
    # "wght=300,400 wdth=75,100" --> {"wght": [300, 400], "wdth": [75, 100]}
    res = {}
    for axis in string:
        try:
            k, v = axis.split("=")
            v = [int(i) for i in v.split(",")]
            res[k] = v
        except ValueError:
            raise ValueError(
                "Incorrect --elided-values input. Requires 'AXIS=val,val ...' "
                "e.g 'wght=400 wdth=100'"
            )
    return res


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "fonts", nargs="+", help="Variable TTF files which make up a family"
    )
    parser.add_argument(
        "--axis-order",
        nargs="+",
        required=True,
        choices=axis_registry.keys(),
        help="List of space seperated axis tags used to set the STAT table "
        "axis order e.g --axis-order wdth wght ital",
    )
    parser.add_argument(
        "--elided-values",
        nargs="+",
        default=None,
        help="List of space seperated axis_values to elide. "
        "Input must be structed as axis_tag=int,int..."
        "e.g --elided-values wdth=100 wght=400",
    )
    parser.add_argument("--out", "-o", help="Output dir for fonts")
    parser.add_argument(
        "--inplace", action="store_true", default=False, help="Overwrite input files"
    )
    args = parser.parse_args()

    fonts = [TTFont(f) for f in args.fonts]
    elided_values = (
        parse_elided_values(args.elided_values) if args.elided_values else None
    )
    gen_stat_tables(fonts, args.axis_order, elided_values)

    if args.out:
        if not os.path.isdir(args.out):
            os.mkdir(args.out)

    for font in fonts:
        if args.out:
            dst = os.path.join(args.out, os.path.basename(font.reader.file.name))
        elif args.inplace:
            dst = font.reader.file.name
        else:
            dst = font.reader.file.name + ".fix"
        if os.path.isfile(dst):
            os.remove(dst)
        print(f"Saving font to {dst}")
        font.save(dst)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
gftools fix-family

Update a family so it conforms to the Google Fonts specification
https://github.com/googlefonts/gf-docs/tree/master/Spec

Usage:

gftools fix-family fonts1.ttf fonts2.ttf

# Fix font issues that should be fixed in the source files
gftools fix-family fonts1.ttf --include-source-fixes
"""
import argparse
import logging
import os
from fontTools.ttLib import TTFont
from gftools.fix import *


log = logging.getLogger(__name__)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("fonts", nargs="+", help="Font family paths")
    parser.add_argument(
        "--inplace", action="store_true", default=False, help="Save fixed fonts inplace"
    )
    parser.add_argument("-o", "--out", help="Output dir for fixed fonts")
    parser.add_argument(
        "--include-source-fixes",
        action="store_true",
        help="Fix font issues that should be fixed in the source files.",
    )
    args = parser.parse_args()

    fonts = [TTFont(f) for f in args.fonts]
    fix_family(fonts, args.include_source_fixes)

    if args.inplace:
        for font in fonts:
            font.save(font.reader.file.name)
    elif args.out:
        if not os.path.isdir(args.out):
            os.mkdir(args.out)
        for font in fonts:
            out_path = os.path.join(
                args.out, os.path.basename(font.reader.file.name)
            )
            font.save(out_path)
    else:
        for font in fonts:
            font.save(font.reader.file.name + ".fix")


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
A Python script to set a font's OS/2 usWeightClass value so it conforms to
the Google Fonts specification. The font's filename is used to determine
the correct value.
"""
from __future__ import print_function
from fontTools.ttLib import TTFont
from fontbakery.parse import style_parse
from gftools.fix import fix_weight_class
import sys
import os


def main(font_path):
    filename = os.path.basename(font_path)
    font = TTFont(font_path)
    current_weight_class = font["OS/2"].usWeightClass
    fix_weight_class(font)
    if current_weight_class != font["OS/2"].usWeightClass:
        font.save(font.reader.file.name + ".fix")
    else:
        print("{}: Skipping. Current WeightClass is correct".format(filename))


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print("Please include a path to a font")
    else:
        main(sys.argv[1])




#!/usr/bin/env python3
"""Tool to take files from a font family project upstream git repository
to the google/fonts GitHub repository structure, taking care of all the details.

Documentation at gftools/docs/gftools-packager/README.md
"""

import sys
from gftools import packager
from gftools.packager import UserAbortError, ProgramAbortError
import argparse

def _ansi_bold(text:str) ->str:
  return f'\033[1m{text}\033[0m'

parser = argparse.ArgumentParser(
    prog='gftools packager',
    description='Package upstream font families for Google Fonts.',
    epilog=f'{_ansi_bold("Documentation:")} '
           'https://github.com/googlefonts/gftools/tree/master/docs/gftools-packager'
           '\n'
           f'{_ansi_bold("Issues:")} '
           'https://github.com/googlefonts/gftools/issues'
)

parser.add_argument(
            'file_or_families',
            metavar='name',
            type=str,
            nargs='*',
            help='The family name(s) or file name(s) of upstream conf yaml '
                 'files to be packaged. If a name ends with the ".yaml" suffix, '
                 'it\'s treated as a file otherwise it\'s used as family name '
                 'and  packager tries to gather upstream configuration from '
                 'the google/fonts GitHub repository. If no name is specified, '
                 'no package will be created. This is useful to only make a '
                 'PR from an already created branch, not adding a commit, '
                 'use -b/--branch and see see -p/--pr.')
parser.add_argument(
            '-f','--force',
            action='store_true',
            help='This allows the program to manipulate/change/delete data '
                 'in [target]. Without this flag only adding new items '
                 '(depends: files, directories or branches, trees, blobs) '
                 'is allowed.')
parser.add_argument(
            '-y',
            '--no-confirm',
            dest='yes',
            action='store_true',
            help='Don\'t require user interaction, by answering with the '
                 'default always. Removes all interactivity.')
parser.add_argument(
            '-q',
            '--quiet',
            action='store_true',
            help='Don\'t print user interaction dialogues when -y/--no-confirm is used.')
parser.add_argument(
            'target',
            type=str,
            help='The target of the package. By default a path to a directory. '
                 'See -f/--force to allow changing none-empty directories. '
                 'See -g/--gf-git to use it as a git repository. '
                 'A notable exception is -u/--upstream-yaml where the upstream.yaml '
                 'template will be saved to target file name.')
parser.add_argument(
            '-g','--gf-git',
            dest='is_gf_git',
            action='store_true',
            help='Try to use target as a git repository clone of GitHub google/fonts and '
                 'create or override a branch from upstream master using a generated '
                 'default branch name or a branch name specified with -b/--branch')
parser.add_argument(
            '-b', '--branch',
            type=str,
            default=None,
            help='Set the local target branch name instead '
                 'of using the generated branch name, like: "gftools_packager_{familyname}". '
                 'This implies -g/--gf-git, i.e. target will be treated as if -g/--gf-git is set.')
parser.add_argument(
            '-a', '--add-commit',
            action='store_true',
            help='Don\'t override existing branch and instead add a new '
                 'commit to the branch. Use this to create a PR for multiple '
                 'familes e.g. a super family or a bunch update. '
                 'It\'s likely that you want to combine this with -b/--branch. '
                 'This implies -g/--gf-git, i.e. target will be treated as if -g/--gf-git is set.')
parser.add_argument(
            '-p', '--pr',
            action='store_true',
            help='Make a pull request. '
                 'This implies -g/--gf-git, i.e. target will be treated as if -g/--gf-git is set. '
                 'See --pr-upstream  and --push-upstream.')
parser.add_argument(
            '--pr-upstream',
            type=str,
            default='',
            help='The upstream where the pull request goes, as a GitHub '
                 '"owner/repoName" pair (default: google/fonts). '
                 'This implies -p/--pr, i.e. target will be treated as if -p/--pr is set.'
                 )
parser.add_argument(
            '--push-upstream',
            type=str,
            default='',
            # we can push to a clone of google/fonts and then pr from
            # that clone to --pr-upstream, however, our ghactions QA can't
            # run on a different repo, that's why this is mostly for testing.
            help='The upstream where the push goes, as a GitHub "owner/repoName" '
                 'pair (default: the value of --pr-upstream). '
                 'This implies -p/--pr, i.e. target will be treated as if -p/--pr is set.')
parser.add_argument(
            '-u', '--upstream-yaml',
            action='store_true',
            help='Create and output the upstream.yaml to the file name given by target. '
                 'This is intended to help bootstrapping new upstream configurations. '
                 'In it\'s simplest form, if no name argument is given, it will output the '
                 'yaml template. '
                 'However, if name is given, this will also try to include all available '
                 'information and interact with the user. This can only handle one name, '
                 'because there can only be one target. '
                 'Use -y/--no-confirm to skip interactive mode.'
                 'Use -f/--force to override existing target.')
parser.add_argument(
            '--no-whitelist',
            action='store_true',
            help='Don\'t use the whitelist of allowed files to copy from '
                 'TARGET in upstream-conf "files". This is meant to enable '
                 'forward compatibility with new files and should not '
                 'be used regularly. Instead file an issue to add new '
                 'files to the whitelist.')
parser.add_argument(
            '--no-source',
            action='store_true',
            help='Don\'t add the "source" key to METADATA.pb. Use this temporarily '
            'until all back-end systems have been updated. '
            'See https://github.com/google/fonts/issues/2587'
            )
parser.add_argument(
            '-B', '--allow-build',
            action='store_true',
            help='Allow executing the bash command in stored in the "build" '
            'key of upstream-conf, if present. Don\'t allow this lightly '
            'and review build command, build process and its dependencies prior. '
            'This support for building from sources is provisional, a '
            'discussion can be found at https://github.com/googlefonts/gftools/issues/231'
            )


if __name__ == '__main__':
    args = parser.parse_args()
    try:
      packager.make_package(**args.__dict__)
    except UserAbortError as e:
      print('Aborted',
            'by user!' if not args.yes else \
            'by program! User interaction required (don\'t use the -y/--no-confirm flag).',
            f'{e}')
      sys.exit(1)
    except ProgramAbortError as e:
      print(f'Aborted by program: {e}')
      sys.exit(1)
    print('Done!')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: axes.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='axes.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\naxes.proto\"\xb5\x01\n\tAxisProto\x12\x0b\n\x03tag\x18\x01 \x01(\t\x12\x14\n\x0c\x64isplay_name\x18\x07 \x01(\t\x12\x11\n\tmin_value\x18\x02 \x01(\x02\x12\x15\n\rdefault_value\x18\x03 \x01(\x02\x12\x11\n\tmax_value\x18\x04 \x01(\x02\x12\x11\n\tprecision\x18\x05 \x01(\x05\x12 \n\x08\x66\x61llback\x18\x06 \x03(\x0b\x32\x0e.FallbackProto\x12\x13\n\x0b\x64\x65scription\x18\x08 \x01(\t\",\n\rFallbackProto\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x02'
)




_AXISPROTO = _descriptor.Descriptor(
  name='AxisProto',
  full_name='AxisProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='tag', full_name='AxisProto.tag', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='display_name', full_name='AxisProto.display_name', index=1,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='min_value', full_name='AxisProto.min_value', index=2,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='default_value', full_name='AxisProto.default_value', index=3,
      number=3, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='max_value', full_name='AxisProto.max_value', index=4,
      number=4, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='precision', full_name='AxisProto.precision', index=5,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fallback', full_name='AxisProto.fallback', index=6,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='AxisProto.description', index=7,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=15,
  serialized_end=196,
)


_FALLBACKPROTO = _descriptor.Descriptor(
  name='FallbackProto',
  full_name='FallbackProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='FallbackProto.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='FallbackProto.value', index=1,
      number=2, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=198,
  serialized_end=242,
)

_AXISPROTO.fields_by_name['fallback'].message_type = _FALLBACKPROTO
DESCRIPTOR.message_types_by_name['AxisProto'] = _AXISPROTO
DESCRIPTOR.message_types_by_name['FallbackProto'] = _FALLBACKPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

AxisProto = _reflection.GeneratedProtocolMessageType('AxisProto', (_message.Message,), {
  'DESCRIPTOR' : _AXISPROTO,
  '__module__' : 'axes_pb2'
  # @@protoc_insertion_point(class_scope:AxisProto)
  })
_sym_db.RegisterMessage(AxisProto)

FallbackProto = _reflection.GeneratedProtocolMessageType('FallbackProto', (_message.Message,), {
  'DESCRIPTOR' : _FALLBACKPROTO,
  '__module__' : 'axes_pb2'
  # @@protoc_insertion_point(class_scope:FallbackProto)
  })
_sym_db.RegisterMessage(FallbackProto)


# @@protoc_insertion_point(module_scope)



from fontTools.ttLib import newTable, TTFont
from gftools.fix import *
from glob import glob
import pytest
import os
from copy import deepcopy


TEST_DATA = os.path.join("data", "test")


@pytest.fixture
def static_font():
    return TTFont(os.path.join(TEST_DATA, "Lora-Regular.ttf"))


@pytest.fixture
def var_font():
    return TTFont(os.path.join(TEST_DATA, "Inconsolata[wdth,wght].ttf"))


@pytest.fixture
def var_fonts():
    paths = [
        os.path.join(TEST_DATA, "Raleway[wght].ttf"),
        os.path.join(TEST_DATA, "Raleway-Italic[wght].ttf")
    ]
    return [TTFont(p) for p in paths]


@pytest.fixture
def static_fonts():
    return [TTFont(f) for f in glob(os.path.join("data", "test", "mavenpro", "*.ttf"))]


def test_remove_tables(static_font):
    # Test removing a table which is part of UNWANTED_TABLES
    tsi1_tbl = newTable("TSI1")
    static_font["TSI1"] = tsi1_tbl
    assert "TSI1" in static_font

    tsi2_tbl = newTable("TSI2")
    static_font["TSI2"] = tsi2_tbl
    remove_tables(static_font, ["TSI1", "TSI2"])
    assert "TSI1" not in static_font
    assert "TSI2" not in static_font

    # Test removing a table which is essential
    remove_tables(static_font, ["glyf"])
    assert "glyf" in static_font


def test_add_dummy_dsig(static_font):
    assert "DSIG" not in static_font
    add_dummy_dsig(static_font)
    assert "DSIG" in static_font


def test_fix_hinted_font(static_font):
    static_font["head"].flags &= ~(1 << 3)
    assert static_font["head"].flags & (1 << 3) != (1 << 3)
    fix_hinted_font(static_font)
    assert static_font["head"].flags & (1 << 3) == (1 << 3)


def test_fix_unhinted_font(static_font):
    for tbl in ("prep", "gasp"):
        if tbl in static_font:
            del static_font[tbl]

    fix_unhinted_font(static_font)
    assert static_font["gasp"].gaspRange == {65535: 15}
    assert "prep" in static_font


def test_fix_fs_type(static_font):
    static_font["OS/2"].fsType = 1
    assert static_font["OS/2"].fsType == 1
    fix_fs_type(static_font)
    assert static_font["OS/2"].fsType == 0


# Taken from https://github.com/googlefonts/gf-docs/tree/master/Spec#supported-styles
STYLE_HEADERS = "style, weight_class, fs_selection, mac_style"
STYLE_TABLE = [
    ("Hairline", 1, (1 << 6), (0 << 0)),
    ("Thin", 100, (1 << 6), (0 << 0)),
    ("ExtraLight", 200, (1 << 6), (0 << 0)),
    ("Light", 300, (1 << 6), (0 << 0)),
    ("Regular", 400, (1 << 6), (0 << 0)),
    ("Medium", 500, (1 << 6), (0 << 0)),
    ("SemiBold", 600, (1 << 6), (0 << 0)),
    ("Bold", 700, (1 << 5), (1 << 0)),
    ("ExtraBold", 800, (1 << 6), (0 << 0)),
    ("Black", 900, (1 << 6), (0 << 0)),
    ("ExtraBlack", 1000, (1 << 6), (0 << 0)),
    ("Hairline Italic", 1, (1 << 0), (1 << 1)),
    ("Thin Italic", 100, (1 << 0), (1 << 1)),
    ("ExtraLight Italic", 200, (1 << 0), (1 << 1)),
    ("Light Italic", 300, (1 << 0), (1 << 1)),
    ("Italic", 400, (1 << 0), (1 << 1)),
    ("Medium Italic", 500, (1 << 0), (1 << 1)),
    ("SemiBold Italic", 600, (1 << 0), (1 << 1)),
    ("Bold Italic", 700, (1 << 0) | (1 << 5), (1 << 0) | (1 << 1)),
    ("ExtraBold Italic", 800, (1 << 0), (1 << 1)),
    ("Black Italic", 900, (1 << 0), (1 << 1)),
    ("ExtraBlack Italic", 1000, (1 << 0), (1 << 1)),
    # Variable fonts may have tokens other than weight and italic in their names
    ("SemiCondensed Bold Italic", 700, (1 << 0) | (1 << 5), (1 << 0) | (1 << 1)),
    ("12pt Italic", 400, (1 << 0), (1 << 1)),
]

@pytest.mark.parametrize(
    STYLE_HEADERS,
    STYLE_TABLE
)
def test_fix_weight_class(static_font, style, weight_class, fs_selection, mac_style):
    name = static_font["name"]
    name.setName(style, 2, 3, 1, 0x409)
    name.setName(style, 17, 3, 1, 0x409)
    fix_weight_class(static_font)
    assert static_font["OS/2"].usWeightClass == weight_class


def test_unknown_weight_class(static_font):
    name = static_font["name"]
    name.setName("Foobar", 2, 3, 1, 0x409)
    name.setName("Foobar", 17, 3, 1, 0x409)
    from gftools.fix import WEIGHT_NAMES

    with pytest.raises(ValueError, match="Cannot determine usWeightClass"):
        fix_weight_class(static_font)


@pytest.mark.parametrize(
    STYLE_HEADERS,
    STYLE_TABLE
)
def test_fs_selection(static_font, style, weight_class, fs_selection, mac_style):
    # disable fsSelection bits above 6
    for i in range(7, 12):
        static_font["OS/2"].fsSelection &= ~(1 << i)
    name = static_font["name"]
    name.setName(style, 2, 3, 1, 0x409)
    name.setName(style, 17, 3, 1, 0x409)
    fix_fs_selection(static_font)
    assert static_font["OS/2"].fsSelection == fs_selection


@pytest.mark.parametrize(
    STYLE_HEADERS,
    STYLE_TABLE
)
def test_fix_mac_style(static_font, style, weight_class, fs_selection, mac_style):
    name = static_font["name"]
    name.setName(style, 2, 3, 1, 0x409)
    name.setName(style, 17, 3, 1, 0x409)
    fix_mac_style(static_font)
    assert static_font["head"].macStyle == mac_style


STYLENAME_HEADERS = "family_name, style, id1, id2, id16, id17"
STYLENAME_TABLE = [
    # Roman
    ("Test Family", "Hairline", "Test Family Hairline", "Regular", "Test Family", "Hairline"),
    ("Test Family", "Thin", "Test Family Thin", "Regular", "Test Family", "Thin"),
    ("Test Family", "ExtraLight", "Test Family ExtraLight", "Regular", "Test Family", "ExtraLight"),
    ("Test Family", "Light", "Test Family Light", "Regular", "Test Family", "Light"),
    ("Test Family", "Regular", "Test Family", "Regular", "", ""),
    ("Test Family", "Medium", "Test Family Medium", "Regular", "Test Family", "Medium"),
    ("Test Family", "SemiBold", "Test Family SemiBold", "Regular", "Test Family", "SemiBold"),
    ("Test Family", "Bold", "Test Family", "Bold", "", ""),
    ("Test Family", "ExtraBold", "Test Family ExtraBold", "Regular", "Test Family", "ExtraBold"),
    ("Test Family", "Black", "Test Family Black", "Regular", "Test Family", "Black"),
    ("Test Family", "ExtraBlack", "Test Family ExtraBlack", "Regular", "Test Family", "ExtraBlack"),
    # Italics
    ("Test Family", "Hairline Italic", "Test Family Hairline", "Italic", "Test Family", "Hairline Italic"),
    ("Test Family", "Thin Italic", "Test Family Thin", "Italic", "Test Family", "Thin Italic"),
    ("Test Family", "ExtraLight Italic", "Test Family ExtraLight", "Italic", "Test Family", "ExtraLight Italic"),
    ("Test Family", "Light Italic", "Test Family Light", "Italic", "Test Family", "Light Italic"),
    ("Test Family", "Italic", "Test Family", "Italic", "", ""),
    ("Test Family", "Medium Italic", "Test Family Medium", "Italic", "Test Family", "Medium Italic"),
    ("Test Family", "SemiBold Italic", "Test Family SemiBold", "Italic", "Test Family", "SemiBold Italic"),
    ("Test Family", "Bold Italic", "Test Family", "Bold Italic", "", ""),
    ("Test Family", "ExtraBold Italic", "Test Family ExtraBold", "Italic", "Test Family", "ExtraBold Italic"),
    ("Test Family", "Black Italic", "Test Family Black", "Italic", "Test Family", "Black Italic"),
    ("Test Family", "ExtraBlack Italic", "Test Family ExtraBlack", "Italic", "Test Family", "ExtraBlack Italic"),
]
@pytest.mark.parametrize(
    STYLENAME_HEADERS,
    STYLENAME_TABLE
)
def test_update_nametable(static_font, family_name, style, id1, id2, id16, id17):
    update_nametable(static_font, family_name, style)
    nametable = static_font["name"]
    assert nametable.getName(1, 3, 1, 0x409).toUnicode() == id1
    assert nametable.getName(2, 3, 1, 0x409).toUnicode() == id2
    if id16 and id17:
        assert nametable.getName(16, 3, 1, 0x409).toUnicode() == id16
        assert nametable.getName(17, 3, 1, 0x409).toUnicode() == id17
    else:
        assert nametable.getName(16, 3, 1, 0x409) == None
        assert nametable.getName(17, 3, 1, 0x409) == None


# TODO test fix_nametable once https://github.com/fonttools/fonttools/pull/2078 is merged


def _get_fvar_instance_names(var_font):
    inst_names = []
    for inst in var_font['fvar'].instances:
        inst_name = var_font['name'].getName(inst.subfamilyNameID, 3, 1, 0x409)
        inst_names.append(inst_name.toUnicode())
    return inst_names


def test_fix_fvar_instances(var_font):
    roman_instances = [
        "ExtraLight",
        "Light",
        "Regular",
        "Medium",
        "SemiBold",
        "Bold",
        "ExtraBold",
        "Black"
    ]
    italic_instances = [
        "ExtraLight Italic",
        "Light Italic",
        "Italic",
        "Medium Italic",
        "SemiBold Italic",
        "Bold Italic",
        "ExtraBold Italic",
        "Black Italic",
    ]
    var_font["fvar"].instances = []

    fix_fvar_instances(var_font)
    inst_names = _get_fvar_instance_names(var_font)
    assert inst_names == roman_instances


    # Let's rename the font style so the font becomes an Italic variant
    var_font2 = deepcopy(var_font)
    var_font2["name"].setName("Italic", 2, 3, 1, 0x409)
    var_font2["name"].setName("Italic", 17, 3, 1, 0x409)

    fix_fvar_instances(var_font2)
    inst_names = _get_fvar_instance_names(var_font2)
    assert inst_names == italic_instances


    # Let's mock an ital axis so the font has both ital and wght axes
    new_fvar = deepcopy(var_font["fvar"])
    new_fvar.axes[1].axisTag = "ital"
    new_fvar.axes[1].minValue = 0
    new_fvar.axes[1].maxValue = 1
    new_fvar.axes[1].defaultValue = 0

    var_font3 = deepcopy(var_font)
    var_font3['fvar'] = new_fvar
    fix_fvar_instances(var_font3)

    inst_names = _get_fvar_instance_names(var_font3)
    assert inst_names == roman_instances + italic_instances


def _check_vertical_metrics(fonts):
    ref_font = fonts[0]
    y_min = min(f["head"].yMin for f in fonts)
    y_max = max(f["head"].yMax for f in fonts)
    for font in fonts:
        # Check fsSelection bit 7 (USE_TYPO_METRICS) is enabled
        assert font["OS/2"].fsSelection & (1 << 7) > 0

        # Check metrics are consistent across family
        assert font["OS/2"].usWinAscent == ref_font["OS/2"].usWinAscent
        assert font["OS/2"].usWinDescent == ref_font["OS/2"].usWinDescent
        assert font["OS/2"].sTypoAscender == ref_font["OS/2"].sTypoAscender
        assert font["OS/2"].sTypoDescender == ref_font["OS/2"].sTypoDescender
        assert font["OS/2"].sTypoLineGap == ref_font["OS/2"].sTypoLineGap
        assert font["hhea"].ascent == ref_font["hhea"].ascent
        assert font["hhea"].descent == ref_font["hhea"].descent
        assert font["hhea"].lineGap == ref_font["hhea"].lineGap

        # Check typo and hhea match
        assert font["OS/2"].sTypoAscender == font["hhea"].ascent
        assert font["OS/2"].sTypoDescender == ref_font["hhea"].descent
        assert font["OS/2"].sTypoLineGap == ref_font["hhea"].lineGap

        # Check win matches family_bounds
        assert font["OS/2"].usWinAscent == y_max
        assert font["OS/2"].usWinDescent == abs(y_min)


def test_fix_vertical_metrics_family_consistency(static_fonts):
    _check_vertical_metrics(static_fonts)
    static_fonts[0]["OS/2"].sTypoLineGap = 1000
    static_fonts[0]["OS/2"].usWinAscent = 4000

    fix_vertical_metrics(static_fonts)
    _check_vertical_metrics(static_fonts)


def test_fix_vertical_metrics_win_values(static_fonts):
    _check_vertical_metrics(static_fonts)
    for font in static_fonts:
        font["OS/2"].usWinAscent = font["OS/2"].usWinDescent = 0
        assert font["OS/2"].usWinAscent == 0 and font["OS/2"].usWinDescent == 0

    fix_vertical_metrics(static_fonts)
    _check_vertical_metrics(static_fonts)


def test_fix_vertical_metrics_typo_and_hhea_match(static_fonts):
    _check_vertical_metrics(static_fonts)
    for font in static_fonts:
        font["hhea"].ascent = 5000
        font["OS/2"].sTypoAscender == 1000
        assert font["hhea"].ascent != font["OS/2"].sTypoAscender

    fix_vertical_metrics(static_fonts)
    _check_vertical_metrics(static_fonts)


def test_fix_vertical_metrics_typo_metrics_enabled(static_fonts):
    _check_vertical_metrics(static_fonts)

    # family currently has fsSelection bit 7 enabled, unset it and change
    # the win Metrics to Typo values
    for font in static_fonts:
        font["OS/2"].fsSelection &= ~(1 << 7)
        font["OS/2"].usWinAscent = 500
        font["OS/2"].usWinDescent = 300

    fix_vertical_metrics(static_fonts)
    # Since fsSelection bit 7 is now enabled, in order for the metrics to visually
    # match the unfixed metrics, the typo values should now be the same as the
    # unfixed win values.
    for font in static_fonts:
        assert font["OS/2"].sTypoAscender == 500
        assert font["OS/2"].sTypoDescender == -300
        assert font["OS/2"].sTypoLineGap == 0
    _check_vertical_metrics(static_fonts)



#!/usr/bin/env python3
# Copyright 2020 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS-IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Check push status of families on Dev, Sandbox and Production servers.

This script will determine whether the families listed in
to_production.txt and to_sandbox.txt in the google/fonts dir have been
pushed to their respective servers from a specific date.

If the files are empty, you can always use git checkout to view past
states.


Usage:

Check push status for the past month:
gftools push-status path/to/google/fonts

Check push status for specified date:
gftools push-status path/to/google/fonts -pd 2020-07-01
"""
import argparse
from argparse import RawDescriptionHelpFormatter
import os
from pathlib import Path
import gftools.fonts_public_pb2 as fonts_pb2
from google.protobuf import text_format
from datetime import datetime, timedelta
import json
import requests


# If no push date is added we can assume that the push has probably
# happened in the past month
ONE_MONTH_AGO = datetime.now() - timedelta(days=31)


def get_family_metadata(url):
    """Get family json data from a Google Fonts metadata url"""
    # can't do requests.get("url").json() since request text starts with ")]}'"
    info = json.loads(requests.get(url).text[5:])
    return {i['family']: i for i in info["familyMetadataList"]}


def read_proto(fp, schema):
    with open(fp, "rb") as f:
        data = text_format.Parse(f.read(), schema)
    return data


def families_from_file(fp):
    """Convert to_sandbox.txt and to_production.txt files to a list of
    family names."""
    results = set()
    with open(fp) as doc:
        family_dirs = [p for p in doc.read().split() if p.startswith(("ofl", "ufl", "apache"))]
    metadata_files = [Path(fp).parent / d / 'METADATA.pb' for d in family_dirs]
    missing_files = [str(f) for f in metadata_files if not f.is_file()]
    if missing_files:
        raise FileNotFoundError(
            "Following METADATA.pbs files are missing:\n{}".format(
                "\n".join(missing_files)
            )
        )
    return [read_proto(f, fonts_pb2.FamilyProto()).name for f in metadata_files]


def families_status(info, push_date, filter_families=set()):
    """Determine which families have been pushed or not pushed from a
    specific date. If a filter_families set is provided, remove
    families from the results which are not included in this set."""
    results = {"pushed": set(), "not_pushed": set()}
    if filter_families:
        info = {k: v for k,v in info.items() if k in filter_families}
    for family in info:
        last_push = info[family]['lastModified']
        last_push = iso_8601_to_date(last_push)
        if last_push >= push_date:
            results['pushed'].add(family)
        if last_push < push_date:
            results["not_pushed"].add(family)
    return results


def status_reporter(server_name, status):
    """Produce a report for a server status dict"""
    report = []
    report.append(f"***{server_name}***")
    for k, v in status.items():
        report.append(f"{k}: {sorted(v)}")
    report.append("\n")
    return "\n".join(report)


def specimen_reporter(server_name, url_base, families):
    """Produce a report containing urls for a list of families"""
    report = []
    report.append(f"***{server_name} specimens to inspect***")
    for family in sorted(families):
        family_url = url_base.format(family.replace(" ", "+"))
        report.append(family_url)
    if len(report) == 1:
        report.append("No urls to inspect")
    report.append("\n")
    return "\n".join(report)


def iso_8601_to_date(string):
    """YYYY-MM-DD --> datetime"""
    if ":" in string or len(string.split("-")) != 3:
        raise ValueError(f"Date format should be 'YYYY-MM-DD' got '{string}'")
    return datetime.strptime(string, "%Y-%m-%d")


def gf_repo(path):
    """Check if path is a Google Fonts repo"""
    path = Path(path)
    if not path.is_dir() or list(path.glob("to_production`.txt")):
        raise OSError(f"{path} is not a Google/Fonts dir")
    return path


def main():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=RawDescriptionHelpFormatter
    )
    parser.add_argument("gf_path", type=gf_repo, help="path to google/fonts dir")
    parser.add_argument(
        "--push_date", "-pd", type=iso_8601_to_date, default=ONE_MONTH_AGO,
        help="Date when last push occurred"
    )
    args = parser.parse_args()

    dev_meta = get_family_metadata("https://fonts-dev.sandbox.google.com/metadata/fonts")
    sandbox_meta = get_family_metadata(f"https://fonts.sandbox.google.com/metadata/fonts")
    prod_meta = get_family_metadata(f"https://fonts.google.com/metadata/fonts")

    dev_status = families_status(dev_meta, args.push_date)

    to_sandbox_file = Path(f"{args.gf_path}/to_sandbox.txt")
    requested_sandbox_families = families_from_file(to_sandbox_file)
    sandbox_status = families_status(sandbox_meta, args.push_date, requested_sandbox_families)

    to_production_file = Path(f"{args.gf_path}/to_production.txt")
    requested_prod_families = families_from_file(to_production_file)
    prod_status = families_status(prod_meta, args.push_date, requested_prod_families)

    specimen_url = "https://fonts-dev.sandbox.google.com/specimen/{}"
    print(specimen_reporter("Dev Server", specimen_url, dev_status["pushed"]))
    print(status_reporter("Sandbox Server", sandbox_status))
    print(status_reporter("Production Server", prod_status))


if __name__ == "__main__":
    main()




from pkg_resources import resource_filename
from gftools.axes_pb2 import AxisProto, FallbackProto
from google.protobuf import text_format
from glob import glob
import os


__all__ = ["axis_registry"]


def AxisRegistry():
    """Parse all axes in the Google Fonts axis registry"""
    results = {}
    axis_reg_dir = resource_filename("gftools", "axisregistry")
    proto_files = glob(os.path.join(axis_reg_dir, "*.textproto"))
    for proto_file in proto_files:
        axis = AxisProto()
        with open(proto_file, "rb") as textproto:
            text_format.Parse(textproto.read(), axis)
            results[axis.tag] = axis
            # Remove spaces from names
            for fallback in results[axis.tag].fallback:
                fallback.name = fallback.name.replace(" ", "")
    return results


axis_registry = AxisRegistry()



"""
Functions to fix fonts so they conform to the Google Fonts
specification:
https://github.com/googlefonts/gf-docs/tree/master/Spec
"""
from fontTools.ttLib import TTFont, newTable
from fontTools.ttLib.tables import ttProgram
from fontTools.ttLib.tables._f_v_a_r import NamedInstance
from gftools.util.google_fonts import _KNOWN_WEIGHTS
from gftools.utils import download_family_from_Google_Fonts, Google_Fonts_has_family
from copy import deepcopy
import logging


log = logging.getLogger(__name__)


__all__ = [
    "remove_tables",
    "add_dummy_dsig",
    "fix_unhinted_font",
    "fix_hinted_font",
    "fix_fs_type",
    "fix_weight_class",
    "fix_fs_selection",
    "fix_mac_style",
    "font_stylename",
    "font_familyname",
    "fix_fvar_instances",
    "update_nametable",
    "fix_nametable",
    "inherit_vertical_metrics",
    "fix_vertical_metrics",
    "fix_font",
    "fix_family",
]


# The _KNOWN_WEIGHT_VALUES constant is used internally by the GF Engineering
# team so we cannot update ourselves. TODO (Marc F) unify this one day
WEIGHT_NAMES = _KNOWN_WEIGHTS
del WEIGHT_NAMES[""]
WEIGHT_NAMES["Hairline"] = 1
WEIGHT_NAMES["ExtraBlack"] = 1000
WEIGHT_VALUES = {v: k for k, v in WEIGHT_NAMES.items()}


UNWANTED_TABLES = frozenset(
    [
        "FFTM",
        "TTFA",
        "TSI0",
        "TSI1",
        "TSI2",
        "TSI3",
        "TSI5",
        "prop",
        "MVAR",
    ]
)


def remove_tables(ttFont, tables=None):
    """Remove unwanted tables from a font. The unwanted tables must belong
    to the UNWANTED_TABLES set.

    Args:
        ttFont: a TTFont instance
        tables: an iterable containing tables remove
    """
    tables_to_remove = UNWANTED_TABLES if not tables else frozenset(tables)
    font_tables = frozenset(ttFont.keys())

    tables_not_in_font = tables_to_remove - font_tables
    if tables_not_in_font:
        log.warning(
            f"Cannot remove tables '{list(tables_not_in_font)}' since they are "
            f"not in the font."
        )

    required_tables = tables_to_remove - UNWANTED_TABLES
    if required_tables:
        log.warning(
            f"Cannot remove tables '{list(required_tables)}' since they are required"
        )

    tables_to_remove = UNWANTED_TABLES & font_tables & tables_to_remove
    if not tables_to_remove:
        return
    log.info(f"Removing tables '{list(tables_to_remove)}' from font")
    for tbl in tables_to_remove:
        del ttFont[tbl]


def add_dummy_dsig(ttFont):
    """Add a dummy dsig table to a font. Older versions of MS Word
    require this table.

    Args:
        ttFont: a TTFont instance
    """
    newDSIG = newTable("DSIG")
    newDSIG.ulVersion = 1
    newDSIG.usFlag = 0
    newDSIG.usNumSigs = 0
    newDSIG.signatureRecords = []
    ttFont.tables["DSIG"] = newDSIG


def fix_unhinted_font(ttFont):
    """Improve the appearance of an unhinted font on Win platforms by:
        - Add a new GASP table with a newtable that has a single
          range which is set to smooth.
        - Add a new prep table which is optimized for unhinted fonts.

    Args:
        ttFont: a TTFont instance
    """
    gasp = newTable("gasp")
    # Set GASP so all sizes are smooth
    gasp.gaspRange = {0xFFFF: 15}

    program = ttProgram.Program()
    assembly = ["PUSHW[]", "511", "SCANCTRL[]", "PUSHB[]", "4", "SCANTYPE[]"]
    program.fromAssembly(assembly)

    prep = newTable("prep")
    prep.program = program

    ttFont["gasp"] = gasp
    ttFont["prep"] = prep


def fix_hinted_font(ttFont):
    """Improve the appearance of a hinted font on Win platforms by enabling
    the head table's flag 3.

    Args:
        ttFont: a TTFont instance
    """
    ttFont["head"].flags |= 1 << 3


def fix_fs_type(ttFont):
    """Set the OS/2 table's fsType flag to 0 (Installable embedding).

    Args:
        ttFont: a TTFont instance
    """
    ttFont["OS/2"].fsType = 0


def fix_weight_class(ttFont):
    """Set the OS/2 table's usWeightClass so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/master/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = stylename.split()
    # Order WEIGHT_NAMES so longest names are first
    for style in sorted(WEIGHT_NAMES, key=lambda k: len(k), reverse=True):
        if style in tokens:
            ttFont["OS/2"].usWeightClass = WEIGHT_NAMES[style]
            return

    if "Italic" in tokens:
        ttFont["OS/2"].usWeightClass = 400
        return
    raise ValueError(
        f"Cannot determine usWeightClass because font style, '{stylename}' "
        f"doesn't have a weight token which is in our known "
        f"weights, '{WEIGHT_NAMES.keys()}'"
    )


def fix_fs_selection(ttFont):
    """Fix the OS/2 table's fsSelection so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/master/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = set(stylename.split())
    fs_selection = ttFont["OS/2"].fsSelection

    # turn off all bits except for bit 7 (USE_TYPO_METRICS)
    fs_selection &= 1 << 7

    if "Italic" in tokens:
        fs_selection |= 1 << 0
    if "Bold" in tokens:
        fs_selection |= 1 << 5
    # enable Regular bit for all other styles
    if not tokens & set(["Bold", "Italic"]):
        fs_selection |= 1 << 6
    ttFont["OS/2"].fsSelection = fs_selection


def fix_mac_style(ttFont):
    """Fix the head table's macStyle so it conforms to GF's supported
    styles table:
    https://github.com/googlefonts/gf-docs/tree/master/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    stylename = font_stylename(ttFont)
    tokens = set(stylename.split())
    mac_style = 0
    if "Italic" in tokens:
        mac_style |= 1 << 1
    if "Bold" in tokens:
        mac_style |= 1 << 0
    ttFont["head"].macStyle = mac_style


def font_stylename(ttFont):
    """Get a font's stylename using the name table. Since our fonts use the
    RIBBI naming model, use the Typographic SubFamily Name (NAmeID 17) if it
    exists, otherwise use the SubFamily Name (NameID 2).

    Args:
        ttFont: a TTFont instance
    """
    return get_name_record(ttFont, 17, fallbackID=2)


def font_familyname(ttFont):
    """Get a font's familyname using the name table. since our fonts use the
    RIBBI naming model, use the Typographic Family Name (NameID 16) if it
    exists, otherwise use the Family Name (Name ID 1).

    Args:
        ttFont: a TTFont instance
    """
    return get_name_record(ttFont, 16, fallbackID=1)


def get_name_record(ttFont, nameID, fallbackID=None, platform=(3, 1, 0x409)):
    """Return a name table record which has the specified nameID.

    Args:
        ttFont: a TTFont instance
        nameID: nameID of name record to return,
        fallbackID: if nameID doesn't exist, use this nameID instead
        platform: Platform of name record. Default is Win US English

    Returns:
        str
    """
    name = ttFont["name"]
    record = name.getName(nameID, 3, 1, 0x409)
    if not record and fallbackID:
        record = name.getName(fallbackID, 3, 1, 0x409)
    if not record:
        raise ValueError(f"Cannot find record with nameID {nameID}")
    return record.toUnicode()


def fix_fvar_instances(ttFont):
    """Replace a variable font's fvar instances with a set of new instances
    that conform to the Google Fonts instance spec:
    https://github.com/googlefonts/gf-docs/tree/master/Spec#fvar-instances

    Args:
        ttFont: a TTFont instance
    """
    if "fvar" not in ttFont:
        raise ValueError("ttFont is not a variable font")

    fvar = ttFont["fvar"]
    default_axis_vals = {a.axisTag: a.defaultValue for a in fvar.axes}

    stylename = font_stylename(ttFont)
    is_italic = "Italic" in stylename
    is_roman_and_italic = any(a for a in ("slnt", "ital") if a in default_axis_vals)

    wght_axis = next((a for a in fvar.axes if a.axisTag == "wght"), None)
    wght_min = int(wght_axis.minValue)
    wght_max = int(wght_axis.maxValue)

    nametable = ttFont["name"]

    def gen_instances(is_italic):
        results = []
        for wght_val in range(wght_min, wght_max + 100, 100):
            name = (
                WEIGHT_VALUES[wght_val]
                if not is_italic
                else f"{WEIGHT_VALUES[wght_val]} Italic".strip()
            )
            name = name.replace("Regular Italic", "Italic")

            coordinates = default_axis_vals
            coordinates["wght"] = wght_val

            inst = NamedInstance()
            inst.subfamilyNameID = nametable.addName(name)
            inst.coordinates = coordinates
            results.append(inst)
        return results

    instances = []
    if is_roman_and_italic:
        for bool_ in (False, True):
            instances += gen_instances(is_italic=bool_)
    elif is_italic:
        instances += gen_instances(is_italic=True)
    else:
        instances += gen_instances(is_italic=False)
    fvar.instances = instances


def update_nametable(ttFont, family_name=None, style_name=None):
    """Update a static font's name table. The updated name table will conform
    to the Google Fonts support styles table:
    https://github.com/googlefonts/gf-docs/tree/master/Spec#supported-styles

    If a style_name includes tokens other than wght and ital, these tokens
    will be appended to the family name e.g

    Input:
    family_name="MyFont"
    style_name="SemiCondensed SemiBold"

    Output:
    familyName (nameID 1) = "MyFont SemiCondensed SemiBold
    subFamilyName (nameID 2) = "Regular"
    typo familyName (nameID 16) = "MyFont SemiCondensed"
    typo subFamilyName (nameID 17) = "SemiBold"

    Google Fonts has used this model for several years e.g
    https://fonts.google.com/?query=cabin

    Args:
        ttFont:
        family_name: New family name
        style_name: New style name
    """
    if "fvar" in ttFont:
        raise ValueError("Cannot update the nametable for a variable font")
    nametable = ttFont["name"]

    # Remove nametable records which are not Win US English
    # TODO this is too greedy. We should preserve multilingual
    # names in the future. Please note, this has always been an issue.
    platforms = set()
    for rec in nametable.names:
        platforms.add((rec.platformID, rec.platEncID, rec.langID))
    platforms_to_remove = platforms ^ set([(3, 1, 0x409)])
    if platforms_to_remove:
        log.warning(
            f"Removing records which are not Win US English, {list(platforms_to_remove)}"
        )
        for platformID, platEncID, langID in platforms_to_remove:
            nametable.removeNames(
                platformID=platformID, platEncID=platEncID, langID=langID
            )

    # Remove any name records which contain linebreaks
    contains_linebreaks = []
    for r in nametable.names:
        for char in ("\n", "\r"):
            if char in r.toUnicode():
                contains_linebreaks.append(r.nameID)
    for nameID in contains_linebreaks:
        nametable.removeNames(nameID)

    if not family_name:
        family_name = font_familyname(ttFont)

    if not style_name:
        style_name = font_stylename(ttFont)

    is_ribbi = style_name in ("Regular", "Bold", "Italic", "Bold Italic")

    nameids = {}
    if is_ribbi:
        nameids[1] = family_name
        nameids[2] = style_name
    else:
        tokens = style_name.split()
        family_name_suffix = " ".join([t for t in tokens if t not in ["Italic"]])
        nameids[1] = f"{family_name} {family_name_suffix}".strip()
        nameids[2] = "Regular" if "Italic" not in tokens else "Italic"

        typo_family_suffix = " ".join(
            t for t in tokens if t not in list(WEIGHT_NAMES) + ["Italic"]
        )
        nameids[16] = f"{family_name} {typo_family_suffix}".strip()
        typo_style = " ".join(t for t in tokens if t in list(WEIGHT_NAMES) + ["Italic"])
        nameids[17] = typo_style

    family_name = nameids.get(16) or nameids.get(1)
    style_name = nameids.get(17) or nameids.get(2)

    # create NameIDs 3, 4, 6
    nameids[4] = f"{family_name} {style_name}"
    nameids[6] = f"{family_name.replace(' ', '')}-{style_name.replace(' ', '')}"
    nameids[3] = _unique_name(ttFont, nameids)

    # Pass through all records and replace occurences of the old family name
    # with the new family name
    current_family_name = font_familyname(ttFont)
    for record in nametable.names:
        string = record.toUnicode()
        if current_family_name in string:
            nametable.setName(
                string.replace(current_family_name, family_name),
                record.nameID,
                record.platformID,
                record.platEncID,
                record.langID,
            )

    # Remove previous typographic names
    for nameID in (16, 17):
        nametable.removeNames(nameID=nameID)

    # Update nametable with new names
    for nameID, string in nameids.items():
        nametable.setName(string, nameID, 3, 1, 0x409)


def _unique_name(ttFont, nameids):
    font_version = _font_version(ttFont)
    vendor = ttFont["OS/2"].achVendID.strip()
    ps_name = nameids[6]
    return f"{font_version};{vendor};{ps_name}"


def _font_version(font, platEncLang=(3, 1, 0x409)):
    nameRecord = font["name"].getName(5, *platEncLang)
    if nameRecord is None:
        return f'{font["head"].fontRevision:.3f}'
    # "Version 1.101; ttfautohint (v1.8.1.43-b0c9)" --> "1.101"
    # Also works fine with inputs "Version 1.101" or "1.101" etc
    versionNumber = nameRecord.toUnicode().split(";")[0]
    return versionNumber.lstrip("Version ").strip()


def fix_nametable(ttFont):
    """Fix a static font's name table so it conforms to the Google Fonts
    supported styles table:
    https://github.com/googlefonts/gf-docs/tree/master/Spec#supported-styles

    Args:
        ttFont: a TTFont instance
    """
    if "fvar" in ttFont:
        # TODO, regen the nametable so it reflects the default fvar axes
        # coordinates. Implement once https://github.com/fonttools/fonttools/pull/2078
        # is merged.
        return
    family_name = font_familyname(ttFont)
    style_name = font_stylename(ttFont)
    update_nametable(ttFont, family_name, style_name)


def _validate_family(ttFonts):
    family_is_vf(ttFonts)
    family_names = set(font_familyname(f) for f in ttFonts)
    if len(family_names) != 1:
        raise ValueError(f"Multiple families found {family_names}")
    return True


def inherit_vertical_metrics(ttFonts, family_name=None):
    """Inherit the vertical metrics from the same family which is
    hosted on Google Fonts.

    Args:
        ttFonts: a list of TTFont instances which belong to a family
        family_name: Optional string which allows users to specify a
            different family to inherit from e.g "Maven Pro".
    """
    family_name = font_familyname(ttFonts[0]) if not family_name else family_name

    gf_fonts = list(map(TTFont, download_family_from_Google_Fonts(family_name)))
    gf_fonts = {font_stylename(f): f for f in gf_fonts}
    # TODO (Marc F) use Regular font instead. If VF use font which has Regular
    # instance
    gf_fallback = list(gf_fonts.values())[0]

    fonts = {font_stylename(f): f for f in ttFonts}
    for style, font in fonts.items():
        if style in gf_fonts:
            src_font = gf_fonts[style]
        else:
            src_font = gf_fallback
        copy_vertical_metrics(src_font, font)

        if typo_metrics_enabled(src_font):
            font["OS/2"].fsSelection |= 1 << 7


def fix_vertical_metrics(ttFonts):
    """Fix a family's vertical metrics based on:
    https://github.com/googlefonts/gf-docs/tree/master/VerticalMetrics

    Args:
        ttFonts: a list of TTFont instances which belong to a family
    """
    src_font = next((f for f in ttFonts if font_stylename(f) == "Regular"), ttFonts[0])

    # TODO (Marc F) CJK Fonts?

    # If OS/2.fsSelection bit 7 isn't enabled, enable it and set the typo metrics
    # to the previous win metrics.
    if not typo_metrics_enabled(src_font):
        src_font["OS/2"].fsSelection |= 1 << 7  # enable USE_TYPO_METRICS
        src_font["OS/2"].sTypoAscender = src_font["OS/2"].usWinAscent
        src_font["OS/2"].sTypoDescender = -src_font["OS/2"].usWinDescent
        src_font["OS/2"].sTypoLineGap = 0

    # Set the hhea metrics so they are the same as the typo
    src_font["hhea"].ascent = src_font["OS/2"].sTypoAscender
    src_font["hhea"].descent = src_font["OS/2"].sTypoDescender
    src_font["hhea"].lineGap = src_font["OS/2"].sTypoLineGap

    # Set the win Ascent and win Descent to match the family's bounding box
    win_desc, win_asc = family_bounding_box(ttFonts)
    src_font["OS/2"].usWinAscent = win_asc
    src_font["OS/2"].usWinDescent = abs(win_desc)

    # Set all fonts vertical metrics so they match the src_font
    for ttFont in ttFonts:
        ttFont["OS/2"].fsSelection |= 1 << 7
        copy_vertical_metrics(src_font, ttFont)


def copy_vertical_metrics(src_font, dst_font):
    for table, key in [
        ("OS/2", "usWinAscent"),
        ("OS/2", "usWinDescent"),
        ("OS/2", "sTypoAscender"),
        ("OS/2", "sTypoDescender"),
        ("OS/2", "sTypoLineGap"),
        ("hhea", "ascent"),
        ("hhea", "descent"),
        ("hhea", "lineGap"),
    ]:
        val = getattr(src_font[table], key)
        setattr(dst_font[table], key, val)


def family_bounding_box(ttFonts):
    y_min = min(f["head"].yMin for f in ttFonts)
    y_max = max(f["head"].yMax for f in ttFonts)
    return y_min, y_max


def typo_metrics_enabled(ttFont):
    return True if ttFont["OS/2"].fsSelection & (1 << 7) else False


def family_is_vf(ttFonts):
    has_fvar = ["fvar" in ttFont for ttFont in ttFonts]
    if any(has_fvar):
        if all(has_fvar):
            return True
        raise ValueError("Families cannot contain both static and variable fonts")
    return False


def fix_italic_angle(ttFont):
    style_name = font_stylename(ttFont)
    if "Italic" not in style_name and ttFont["post"].italicAngle != 0:
        ttFont["post"].italicAngle = 0
    # TODO (Marc F) implement for italic fonts


def fix_font(font, include_source_fixes=False):
    font["OS/2"].version = 4
    if "DSIG" not in font:
        add_dummy_dsig(font)

    if "fpgm" in font:
        fix_hinted_font(font)
    else:
        fix_unhinted_font(font)

    if "fvar" in font:
        remove_tables(font, ["MVAR"])

    if include_source_fixes:
        log.warning(
            "include-source-fixes is enabled. Please consider fixing the "
            "source files instead."
        )
        remove_tables(font)
        fix_nametable(font)
        fix_fs_type(font)
        fix_fs_selection(font)
        fix_mac_style(font)
        fix_weight_class(font)
        fix_italic_angle(font)

        if "fvar" in font:
            fix_fvar_instances(font)
            # TODO (Marc F) add gen-stat once merged
            # https://github.com/googlefonts/gftools/pull/263


def fix_family(fonts, include_source_fixes=False):
    """Fix all fonts in a family"""
    _validate_family(fonts)
    family_name = font_familyname(fonts[0])

    for font in fonts:
        fix_font(font, include_source_fixes=include_source_fixes)

    if include_source_fixes:
        try:
            if Google_Fonts_has_family(family_name):
                inherit_vertical_metrics(fonts)
            else:
                log.warning(
                    f"{family_name} is not on Google Fonts. Skipping "
                    "regression fixes"
                )
        except FileNotFoundError:
            log.warning(
                f"Google Fonts api key not found so we can't regression "
                "fix fonts. See Repo readme to add keys."
            )
        fix_vertical_metrics(fonts)



"""Generate STAT tables for a Variable Font Family

This module exports the function "gen_stat_tables" which can be used to
generate a STAT table for each font in a Variable Font Family.

The STAT AxisValues are constructed using the Google Font's Axis Registry,
https://github.com/google/fonts/tree/master/axisregistry

The function should be able to make STAT tables for any family with the
following properties:
- All fonts contain the same amount of fvar axes
- All fvar axes have the same ranges
"""
from fontTools.otlLib.builder import buildStatTable
from gftools.fix import font_stylename, font_familyname
from gftools.axisreg import axis_registry
import logging


__all__ = ["gen_stat_tables", "ELIDABLE_AXIS_VALUE_NAME"]


log = logging.getLogger(__name__)


ELIDABLE_AXIS_VALUE_NAME = 0x2


# TODO we may have more of these. Please note that some applications may not
# implement variable font style linking.
LINKED_VALUES = {
    "wght": (400, 700),
    "ital": (0.0, 1.0),
}


def _gen_stat_from_fvar(ttFont, axis_reg=axis_registry):
    """Generate a STAT table using a ttFont's fvar and the GF axis registry.

    Args:
        axis_reg: gf axis registry
        ttFont: a TTFont instance
    """
    fvar = ttFont["fvar"]

    axis_defaults = {a.axisTag: a.defaultValue for a in fvar.axes}
    results = {}
    for axis in fvar.axes:
        axis_tag = axis.axisTag
        if axis_tag not in axis_reg:
            log.warning(
                f"'{axis_tag}' isn't in our axis registry. Please open an issue "
                "to discuss the inclusion of this axis, "
                "https://github.com/google/fonts/issues"
            )
            continue
        # Add Axis Record
        results[axis_tag] = {
            "tag": axis_tag,
            "name": axis_reg[axis_tag].display_name,
            "values": [],
        }
        # Add Axis Values
        min_value = axis.minValue
        max_value = axis.maxValue
        for fallback in axis_reg[axis_tag].fallback:
            if fallback.value >= min_value and fallback.value <= max_value:
                axis_value = _add_axis_value(fallback.name, fallback.value)
                # set elided axis_values
                # if axis is opsz, we want to use the fvar default
                if axis_tag == "opsz" and axis_defaults[axis_tag] == fallback.value:
                    axis_value["flags"] |= ELIDABLE_AXIS_VALUE_NAME
                # for all other weights, we want to use the axis reg default
                elif fallback.value == axis_reg[axis_tag].default_value:
                    axis_value["flags"] |= ELIDABLE_AXIS_VALUE_NAME
                results[axis_tag]["values"].append(axis_value)
    return results


def _axes_in_family_name_records(ttFonts):
    results = set()
    for ttFont in ttFonts:
        familyname = font_familyname(ttFont)
        stylename = font_stylename(ttFont)
        results |= set(stylename_to_axes(familyname)) | set(
            stylename_to_axes(stylename)
        )
    return results


def stylename_to_axes(font_style, axis_reg=axis_registry):
    """Get axis names for stylename particles using the axis registry e.g

    "Condensed Bold Italic" --> ["wdth", "wght", "ital"]

    Args:
        axis_reg: gf axis registry
        font_style: str

    Returns: list(str,...)
    """
    axes = []
    unparsed_tokens = []

    tokens = font_style.split()
    for token in tokens:
        axis = style_token_to_axis(token)
        if axis:
            axes.append(axis)
        else:
            unparsed_tokens.append(token)

    if unparsed_tokens:
        log.debug(
            f"Following tokens were not found in the Axis Registry "
            f"{list(unparsed_tokens)}. Axis Values will not be created "
            f"for these tokens"
        )
    return axes


def style_token_to_axis(string, axis_reg=axis_registry):
    # Condensed --> width
    for axis_tag, axis in axis_reg.items():
        for fallback in axis.fallback:
            if fallback.name == string:
                return axis_tag
    return None


def _append_non_fvar_axes_to_stat(
    ttFont, stat_table, axes_in_family_name_records, axis_reg=axis_registry
):
    stylename = font_stylename(ttFont)
    familyname = font_familyname(ttFont)
    style = f"{familyname} {stylename}"
    # {"wght": "Regular", "ital": "Roman", ...}
    font_axes_in_namerecords = {style_token_to_axis(t): t for t in style.split()}

    # Add axes to ttFont which exist across the family but are not in the
    # ttFont's fvar
    axes_missing = axes_in_family_name_records - set(stat_table)
    for axis in axes_missing:
        axis_record = {
            "tag": axis,
            "name": axis_reg[axis].display_name,
            "values": [],
        }
        # Add Axis Value for axis which isn't in the fvar or ttFont style
        # name/family name
        if axis not in font_axes_in_namerecords:
            axis_record["values"].append(_default_axis_value(axis, axis_reg))
        # Add Axis Value for axis which isn't in the fvar but does exist in
        # the ttFont style name/family name
        else:
            style_name = font_axes_in_namerecords[axis]
            value = next(
                (i.value for i in axis_reg[axis].fallback if i.name == style_name),
                None,
            )
            axis_value = _add_axis_value(style_name, value)
            axis_record["values"].append(axis_value)
        stat_table[axis] = axis_record
    return stat_table


def _seen_axis_values(stat_tables):
    seen_axis_values = {}
    for stat_tbl in stat_tables:
        for axis_tag, axis in stat_tbl.items():
            if axis_tag not in seen_axis_values:
                seen_axis_values[axis_tag] = set()
            seen_axis_values[axis_tag] |= set(i["value"] for i in axis["values"])
    return seen_axis_values


def _add_linked_axis_values_to_stat(stat_table, seen_axis_values):
    for axis_tag, axis in stat_table.items():
        if axis_tag not in LINKED_VALUES:
            continue
        start, end = LINKED_VALUES[axis_tag]
        for axis_value in axis["values"]:
            if axis_value["value"] == start and end in seen_axis_values[axis_tag]:
                axis_value["linkedValue"] = end
    return stat_table


def _add_elided_axis_values_to_stat(stat_table, elided_values):
    """Overwrite which Axis Values should be elided.

    Args:
        stat: a stat table
        elided_values: dict structured as {"axisTag": [100,200 ...]}"""
    for axis_tag, axis in stat_table.items():
        if axis_tag not in elided_values:
            continue
        for val in axis["values"]:
            if val["value"] in elided_values[axis_tag]:
                val["flags"] |= ELIDABLE_AXIS_VALUE_NAME
            else:
                val["flags"] &= ~ELIDABLE_AXIS_VALUE_NAME
    return stat_table


def _add_axis_value(style_name, value, flags=0x0, linked_value=None):
    value = {"value": value, "name": style_name, "flags": flags}
    if linked_value:
        value["linkedValue"] = linked_value
    return value


def _default_axis_value(axis, axis_reg=axis_registry):
    axis_record = axis_reg[axis]
    default_value = axis_record.default_value
    default_name = next(
        (i.name for i in axis_record.fallback if i.value == default_value), None
    )
    return _add_axis_value(default_name, default_value, flags=ELIDABLE_AXIS_VALUE_NAME)


def validate_axis_order(axis_order, seen_axes):
    axes_not_ordered = seen_axes - set(axis_order)
    if axes_not_ordered:
        raise ValueError(f"Axis order arg is missing {axes_not_ordered} axes.")


def validate_family_fvar_tables(ttFonts):
    """Google Fonts requires all VFs in a family to have the same
    amount of fvar axes and each fvar axis should have the same range.

    Args:
        ttFonts: an iterable containing TTFont instances
    """
    for ttFont in ttFonts:
        if "fvar" not in ttFont:
            raise ValueError(f"Font is missing fvar table")

    failed = False
    src_fvar = ttFonts[0]["fvar"]
    src_axes = {a.axisTag: a.__dict__ for a in src_fvar.axes}
    for ttFont in ttFonts:
        fvar = ttFont["fvar"]
        axes = {a.axisTag: a.__dict__ for a in fvar.axes}
        if len(axes) != len(src_axes):
            failed = True
            break
        for axis_tag in axes:
            if axes[axis_tag]["minValue"] != src_axes[axis_tag]["minValue"]:
                failed = True
            if axes[axis_tag]["maxValue"] != src_axes[axis_tag]["maxValue"]:
                failed = True
            # TODO should this fail if default values are different?
    if failed:
        raise ValueError("fvar axes are not consistent across the family")


def _update_fvar_nametable_records(ttFont, stat_table):
    """Add postscript names to fvar instances and add nameID 25 to a
    font's nametable"""
    nametable = ttFont["name"]
    fvar = ttFont["fvar"]
    family_name = font_familyname(ttFont)
    axes_with_one_axis_value = [
        a["values"][0] for a in stat_table if len(a["values"]) == 1
    ]
    tokens = [v["name"] for v in axes_with_one_axis_value]
    tokens = [t for t in tokens if t not in family_name.split()]
    ps_tokens = "".join(t for t in tokens)

    # Variations PostScript Name Prefix
    ps_prefix = f"{family_name}{ps_tokens}".replace(" ", "")
    for rec in [(25, 1, 0, 0), (25, 3, 1, 0x409)]:
        nametable.setName(ps_prefix, *rec)

    # Add or update fvar instance postscript names
    for instance in fvar.instances:
        subfamily_id = instance.subfamilyNameID
        subfamily_name = nametable.getName(subfamily_id, 3, 1, 0x409).toUnicode()
        for token in tokens:
            subfamily_name = subfamily_name.replace(token, "")
            if subfamily_name == "":
                subfamily_name = "Regular"
        ps_name = f"{ps_prefix}-{subfamily_name}".replace(" ", "")
        # Remove ps name records if they already exist
        if instance.postscriptNameID != 65535:
            nametable.removeNames(nameID=instance.postscriptNameID)
        instance.postscriptNameID = nametable.addName(ps_name)


def gen_stat_tables(
    ttFonts, axis_order, elided_axis_values=None, axis_reg=axis_registry
):
    """
    Generate a stat table for each font in a family using the Google Fonts
    Axis Registry.

    Args:
        ttFonts: an iterable containing ttFont instances
        axis_order: a list containing the axis order
        elided_axis_values: a dict containing axes and their values to elide
        e.g {"wght": [400], "wdth": [100]}
        axis_reg: Google Fonts axis registry
    """
    # Heuristic:
    # 1. Gen a STAT table for each font using their fvar tables only
    # 2. Collect all the axes which exist in every font's family name and
    #    and style name
    # 3. Add further Axis Records to each font's stat table for the axes we
    #    found in step 2. Only add them if the stat table doesn't contain them
    #    already.
    # 4. Add an AxisValue to each of the Axes Records we added in step 3.
    #    For each axis in each font, do the following:
    #      a. If a font's name table contains the axis and it is not in the
    #         fvar, we will create a new Axis Value using the axis registry
    #         fallbacks.
    #      b. If a font's name table doesn't contain the axis, we will create a
    #         new Axis Value based the default values found in the axis registry
    #
    #         Example:
    #
    #            Test Case:
    #            Axes in family names: ["wdth", "wght", "ital"]
    #            Font StyleName = "Condensed Bold"
    #            Axes in font fvar = ["wght"]
    #
    #            a result:
    #            axisValue = {"name": "Condensed", "value": 75.0}
    #            "Condensed" exists in our axis registry as a fallback in the
    #            wdth axis
    #
    #            b result:
    #            AxisValue = {"name": "Roman", "value": 0.0, flags=0x2}
    #            Since there isn't an ital token in the Font family name or
    #            style name, the AxisValue will be based on the default values
    #            for the axis in our axis registry
    #
    # 4. For each stat table, update Axis Values which should be linked
    # 5. For each stat table, update Axis Values which should be elided based
    #    on the user arg elided_axis_values (optional)
    # 6. For each stat table, sort axes based on the arg axis_order
    # 7. Use fontTools to build each stat table for each font
    validate_family_fvar_tables(ttFonts)
    stat_tables = [_gen_stat_from_fvar(f) for f in ttFonts]
    axes_in_family_name_records = _axes_in_family_name_records(ttFonts)
    stat_tables = [
        _append_non_fvar_axes_to_stat(ttFont, stat, axes_in_family_name_records)
        for ttFont, stat in zip(ttFonts, stat_tables)
    ]
    seen_axis_values = _seen_axis_values(stat_tables)
    stat_tables = [
        _add_linked_axis_values_to_stat(s, seen_axis_values) for s in stat_tables
    ]
    if elided_axis_values:
        stat_tables = [
            _add_elided_axis_values_to_stat(s, elided_axis_values) for s in stat_tables
        ]

    # TODO make axis_order an optional arg. We can only do this once we
    # have established an axis order in the axis registry
    validate_axis_order(axis_order, set(seen_axis_values.keys()))
    assert len(stat_tables) == len(ttFonts)
    axis_order = [a for a in axis_order if a in seen_axis_values.keys()]
    for stat_table, ttFont in zip(stat_tables, ttFonts):
        stat_table = [stat_table[axis] for axis in axis_order]
        _update_fvar_nametable_records(ttFont, stat_table)
        buildStatTable(ttFont, stat_table)
